package.path = package.path .. ";./?.lua;./?/init.lua;./init.lua"

local HooECS = require('HooECS')
HooECS.initialize({ globals = true, debugging = true })

describe('Entity', function()
    local TestComponent, TestComponent1, TestComponent2, TestComponent3
    local entity, entity1, entity2, parent
    local testSystem, engine

    setup(
    function()
        TestComponent = Component.create('TestComponent')
        TestComponent1 = Component.create('TestComponent1')
        TestComponent2 = Component.create('TestComponent2')
        TestComponent3 = Component.create('TestComponent3')
        CopyTestComponent = Component.create('CopyTestComponent')
        function CopyTestComponent:initialize()
            self.num = 6
        end
    end
    )

    before_each(
    function()
        entity = Entity()
        entity.id = 1
        entity1 = Entity()
        entity1.id = 2
        parent = Entity()
        testComponent = TestComponent()
        testComponent1 = TestComponent1()
        testComponent2 = TestComponent2()
        testComponent3 = TestComponent3()
    end
    )

    it(':add() adds a Component', function()
        entity:add(testComponent)
        assert.are.equal(entity.components[testComponent.class.name], testComponent)
    end)

    it(':add() doesn`t add the same Component twice', function()
        testComponent.int = 12
        entity:add(testComponent)
        assert.are.equal(entity.components[testComponent.class.name].int, 12)
        -- Creation of new testComponent with varying variables
        testComponent = TestComponent()
        testComponent.int = 13
        entity:add(testComponent)
        assert.are_not.equal(entity.components[testComponent.class.name].int, 13)
    end)

    it(':remove() removes a Component', function()
        entity:add(testComponent)
        entity:remove('TestComponent')
    end)

    it(':remove() prints debug message if Component does not exist', function()
        local debug_spy = spy.on(HooECS, 'debug')
        entity:remove('TestComponent')
        assert.spy(debug_spy).was_called()
        HooECS.debug:revert()
    end)

    it(':get() gets a Component', function()
        entity:add(testComponent)
        assert.are.equal(entity:get(testComponent.class.name), testComponent)
    end)

    it(':getComponents() gets all components of an entity', function()
        entity:add(testComponent)
        entity:add(testComponent1)
        components = entity:getComponents()

        local count = 0
        for _, __ in pairs(components) do
            count = count + 1
        end

        assert.True(count == 2)
    end)

    it(':has() shows if it has a Component', function()
        entity:add(testComponent)
        assert.is_true(entity:has(testComponent.class.name))
    end)

    it(':set() adds and overwrites Components', function()
        testComponent.int = 12
        entity:set(testComponent)
        assert.are.equal(entity.components[testComponent.class.name].int, 12)
        testComponent = TestComponent()
        testComponent.int = 13
        entity:set(testComponent)
        assert.are.equal(entity.components[testComponent.class.name].int, 13)
    end)

    it(':addMultiple() adds Multiple Components at once', function()
        local componentList = {testComponent1, testComponent2, testComponent3}
        entity:addMultiple(componentList)
        assert.are.equal(entity.components[testComponent1.class.name], testComponent1)
        assert.are.equal(entity.components[testComponent2.class.name], testComponent2)
        assert.are.equal(entity.components[testComponent3.class.name], testComponent3)
    end)

    it('Constructor with parrent adds a Parent', function()
        entity = Entity(parent)
        assert.are.equal(entity.parent, parent)
    end)

    it(':setParent() adds a Parent', function()
        entity:setParent(parent)
        assert.are.equal(entity.parent, parent)
    end)

    it(':getParent() gets a Parent', function()
        entity:setParent(parent)
        assert.are.equal(entity:getParent(), parent)
    end)

    it(':registerAsChild() registers as a Child', function()
        entity:setParent(parent)
        assert.are.equal(entity:getParent().children[entity.id], entity)
    end)

    it(':copy() creates a deep copy of the entity', function()
        testComponent.num = 42
        entity:add(testComponent)
        local newEntity = entity:copy()
        newEntity:get("TestComponent").num = 21
        assert.are.equal(entity:get("TestComponent").num, 42)
        assert.are.equal(newEntity:get("TestComponent").num, 21)
    end)

    it(':copy(components) creates a shallow copy of the entity replacing components with the provided list', function()
        local tempComponent = TestComponent1()
        tempComponent.num = 1
        testComponent1 = 2
        testComponent.num = 42
        entity:add(testComponent)
        local newEntity = entity:copy({tempComponent})
        newEntity:get("TestComponent").num = 21
        assert.are.equal(entity:get("TestComponent").num, 42)
        assert.are.equal(newEntity:get("TestComponent").num, 21)
    end)

    it(':shallowCopy() creates a shallow copy of the entity', function()
        testComponent.num = 42
        entity:add(testComponent)
        local newEntity = entity:shallowCopy()
        newEntity:get("TestComponent").num = 21
        assert.are.equal(entity:get("TestComponent").num, 21)
        assert.are.equal(newEntity:get("TestComponent").num, 21)
    end)

    it(':shallowCopy(components) creates a shallow copy of the entity replacing components with the provided list', function()
        local tempComponent = TestComponent1()
        tempComponent.num = 1
        testComponent1.num = 2
        testComponent.num = 42
        entity:add(testComponent)
        entity:add(testComponent1)
        local newEntity = entity:shallowCopy({tempComponent})
        newEntity:get("TestComponent").num = 21
        assert.are.equal(entity:get("TestComponent").num, 21)
        assert.are.equal(newEntity:get("TestComponent").num, 21)
        assert.are.equal(newEntity:get("TestComponent1").num, 1)
    end)

    it('calls Component:removedFromEntity and Component:addedToEntity when appropriate', function()
        local tempComponent = TestComponent()
        testComponent.addedToEntity = function(self, entity)
            entity.num = 1
            self.otherNum = 1
        end
        tempComponent.addedToEntity = function (self, entity)
            entity.num = 2
            self.otherNum = 2
        end
        tempComponent.removedFromEntity = function(self, entity)
            entity.num = 3
            self.otherNum = 3
        end

        assert.are.equal(entity.num, nil)
        assert.are.equal(testComponent.otherNum, nil)

        entity:add(testComponent)
        assert.are.equal(entity.num, 1)
        assert.are.equal(testComponent.otherNum, 1)

        entity:set(tempComponent)
        assert.are.equal(entity.num, 2)
        assert.are.equal(tempComponent.otherNum, 2)

        entity:remove("TestComponent")
        assert.are.equal(entity.num, 3)
        assert.are.equal(tempComponent.otherNum, 3)
    end)

    it(':getChildren() returns children or nil if non exit', function()
        assert.are.equal(entity:getChildren(), nil)
        entity1:setParent(entity)
        assert.are.equal(type(entity:getChildren()), "table")
        local k, v = next(entity:getChildren())
        assert.are.equal(v, entity1)
    end)
end)
